// src/writer.ts
var Writer = class {
  write(operation) {
    throw new Error("Method 'write()' must be implemented.");
  }
  close() {
    throw new Error("Method 'close()' must be implemented.");
  }
};

// src/array.ts
var ArrayWriter = class extends Writer {
  operations = [];
  write(operation) {
    this.operations.push(operation);
  }
  close() {
    return this.operations;
  }
};
function diffArrayToString(diffArray, options = {}) {
  let omit = options.omit ?? [];
  let insertTagOpen = options.insertTagOpen ?? "[+ ";
  let insertTagClose = options.insertTagClose ?? " +]";
  let deleteTagOpen = options.deleteTagOpen ?? "[- ";
  let deleteTagClose = options.deleteTagClose ?? " -]";
  let equalTagOpen = options.equalTagOpen ?? "";
  let equalTagClose = options.equalTagClose ?? "";
  let result = "";
  for (const change of diffArray) {
    if (change.type === "insert" && !omit.includes("insert")) {
      result += insertTagOpen;
      for (const token of change.tokens) {
        result += token.value;
      }
      result += insertTagClose;
    }
    if (change.type === "delete" && !omit.includes("delete")) {
      result += deleteTagOpen;
      for (const token of change.tokens) {
        result += token.value;
      }
      result += deleteTagClose;
    }
    if (change.type === "equal" && !omit.includes("equal")) {
      result += equalTagOpen;
      for (const token of change.tokens) {
        result += token.value;
      }
      result += equalTagClose;
    }
  }
  return result;
}
function diffStringToArray(diffString, options = {}) {
  const insertTagOpen = options.insertTagOpen ?? "[+ ";
  const insertTagClose = options.insertTagClose ?? " +]";
  const deleteTagOpen = options.deleteTagOpen ?? "[- ";
  const deleteTagClose = options.deleteTagClose ?? " -]";
  const equalTagOpen = options.equalTagOpen ?? "";
  const equalTagClose = options.equalTagClose ?? "";
  const operations = [];
  let currentIndex = 0;
  const lines = diffString.split("\n");
  const hasUnifiedDiffMarkers = lines.some(
    (line) => (line.startsWith("+") || line.startsWith("-")) && !line.startsWith("++") && !line.startsWith("--")
  );
  if (hasUnifiedDiffMarkers) {
    for (const line of lines) {
      if (line.startsWith("@@")) {
        currentIndex += line.length + 1;
        continue;
      }
      if (line.startsWith("+")) {
        operations.push({
          type: "insert",
          tokens: [
            {
              value: line.slice(1) + "\n",
              start: currentIndex,
              end: currentIndex + line.length
            }
          ]
        });
      } else if (line.startsWith("-")) {
        operations.push({
          type: "delete",
          tokens: [
            {
              value: line.slice(1) + "\n",
              start: currentIndex,
              end: currentIndex + line.length
            }
          ]
        });
      } else {
        const value = line.startsWith(" ") ? line.slice(1) : line;
        operations.push({
          type: "equal",
          tokens: [
            {
              value: value + "\n",
              start: currentIndex,
              end: currentIndex + line.length
            }
          ]
        });
      }
      currentIndex += line.length + 1;
    }
    return operations;
  }
  while (currentIndex < diffString.length) {
    if (diffString.startsWith(insertTagOpen, currentIndex)) {
      const contentStart = currentIndex + insertTagOpen.length;
      const contentEnd = diffString.indexOf(insertTagClose, contentStart);
      if (contentEnd === -1) throw new Error("Missing insert tag close");
      const content = diffString.slice(contentStart, contentEnd);
      operations.push({
        type: "insert",
        tokens: [{ value: content, start: contentStart, end: contentEnd }]
      });
      currentIndex = contentEnd + insertTagClose.length;
    } else if (diffString.startsWith(deleteTagOpen, currentIndex)) {
      const contentStart = currentIndex + deleteTagOpen.length;
      const contentEnd = diffString.indexOf(deleteTagClose, contentStart);
      if (contentEnd === -1) throw new Error("Missing delete tag close");
      const content = diffString.slice(contentStart, contentEnd);
      operations.push({
        type: "delete",
        tokens: [{ value: content, start: contentStart, end: contentEnd }]
      });
      currentIndex = contentEnd + deleteTagClose.length;
    } else {
      let contentEnd;
      if (equalTagOpen === "" && equalTagClose === "") {
        const nextInsert = diffString.indexOf(insertTagOpen, currentIndex);
        const nextDelete = diffString.indexOf(deleteTagOpen, currentIndex);
        contentEnd = Math.min(
          nextInsert === -1 ? Infinity : nextInsert,
          nextDelete === -1 ? Infinity : nextDelete
        );
        if (contentEnd === Infinity) contentEnd = diffString.length;
      } else {
        const contentStart = currentIndex + equalTagOpen.length;
        contentEnd = diffString.indexOf(equalTagClose, contentStart);
        if (contentEnd === -1) throw new Error("Missing equal tag close");
        contentEnd += equalTagClose.length;
      }
      if (contentEnd > currentIndex) {
        const content = diffString.slice(
          currentIndex + equalTagOpen.length,
          contentEnd - (equalTagClose.length || 0)
        );
        if (content) {
          operations.push({
            type: "equal",
            tokens: [{ value: content, start: currentIndex, end: contentEnd }]
          });
        }
      }
      currentIndex = contentEnd;
    }
  }
  return operations;
}

// src/diff.ts
function diffTokens({
  a,
  b,
  writer
}) {
  if (!writer) {
    writer = new ArrayWriter();
  }
  const hashCache = /* @__PURE__ */ new Map();
  const skipTokens = [";", ","];
  const lcsMatrix = computeLCSMatrix(a.tokens, b.tokens);
  let i = 0;
  let j = 0;
  let whileLimit = Math.max(
    a.tokens.length + b.tokens.length,
    a.tokens.length * b.tokens.length
  ) * 2;
  let operation = { type: "equal", tokens: [] };
  while ((i < a.tokens.length || j < b.tokens.length) && whileLimit-- >= 0) {
    if (i < a.tokens.length && skipTokens.includes(a.tokens[i].value)) {
      i++;
      continue;
    }
    if (j < b.tokens.length && skipTokens.includes(b.tokens[j].value)) {
      if (operation.type !== "equal") {
        writer.write(operation);
        operation = { type: "equal", tokens: [] };
      }
      operation.tokens.push(b.tokens[j]);
      j++;
      continue;
    }
    const aToken = i < a.tokens.length ? {
      ...a.tokens[i],
      value: a.tokens[i].value.replaceAll(
        a.indentType.repeat(a.indentAmount),
        b.indentType.repeat(b.indentAmount)
      )
    } : null;
    const bToken = j < b.tokens.length ? b.tokens[j] : null;
    if (!aToken) {
      if (operation.type !== "insert") {
        writer.write(operation);
        operation = { type: "insert", tokens: [] };
      }
      if (bToken) {
        operation.tokens.push(bToken);
        j++;
      }
      continue;
    }
    if (!bToken) {
      if (operation.type !== "delete") {
        writer.write(operation);
        operation = { type: "delete", tokens: [] };
      }
      if (aToken) {
        operation.tokens.push(aToken);
        i++;
      }
      continue;
    }
    const aValue = hashToken(aToken.value);
    const bValue = hashToken(bToken.value);
    if (aValue === bValue && lcsMatrix[i][j] > lcsMatrix[i + 1][j]) {
      if (operation.type !== "equal") {
        writer.write(operation);
        operation = { type: "equal", tokens: [] };
      }
      operation.tokens.push(bToken);
      i++;
      j++;
      continue;
    }
    if (bToken && (i >= a.tokens.length || lcsMatrix[i][j + 1] >= lcsMatrix[i + 1][j])) {
      if (operation.type !== "insert") {
        writer.write(operation);
        operation = { type: "insert", tokens: [] };
      }
      operation.tokens.push(bToken);
      j++;
      continue;
    }
    if (aToken) {
      if (operation.type !== "delete") {
        writer.write(operation);
        operation = { type: "delete", tokens: [] };
      }
      operation.tokens.push(aToken);
      i++;
    }
  }
  if (operation.tokens.length > 0) {
    writer.write(operation);
    operation = { type: "equal", tokens: [] };
  }
  if (whileLimit <= 0) {
    throw new Error("while loop timeout");
  }
  return writer.close();
  function computeLCSMatrix(tokensA, tokensB) {
    const timeStart = performance.now();
    const WINDOW_SIZE = 200;
    const lcsMatrix2 = Array(tokensA.length + 1).fill(null).map(() => Array(tokensB.length + 1).fill(0));
    for (let i2 = tokensA.length - 1; i2 >= 0; i2--) {
      const windowStart = Math.max(0, i2 - WINDOW_SIZE);
      const windowEnd = Math.min(tokensB.length, i2 + WINDOW_SIZE);
      for (let j2 = windowEnd - 1; j2 >= windowStart; j2--) {
        const aValue = hashToken(tokensA[i2].value);
        const bValue = hashToken(tokensB[j2].value);
        if (aValue === bValue) {
          lcsMatrix2[i2][j2] = lcsMatrix2[i2 + 1][j2 + 1] + 1;
        } else {
          lcsMatrix2[i2][j2] = Math.max(lcsMatrix2[i2 + 1][j2], lcsMatrix2[i2][j2 + 1]);
        }
      }
    }
    return lcsMatrix2;
  }
  function hashToken(token) {
    if (hashCache.has(token)) {
      return hashCache.get(token);
    }
    const result = [];
    let lastWasSpace = false;
    for (let i2 = 0; i2 < token.length; i2++) {
      const c = token[i2];
      if (c === " " || c === "	" || c === "\n" || c === "\r" || c === "\f" || c === "\v") {
        if (!lastWasSpace) {
          result.push(" ");
          lastWasSpace = true;
        }
      } else if (c === '"' || c === "'" || c === "`") {
        result.push('"');
        lastWasSpace = false;
      } else {
        result.push(c);
        lastWasSpace = false;
      }
    }
    const hash = result.join("");
    hashCache.set(token, hash);
    return hash;
  }
}

// src/indents.ts
var INDENT_REGEX = /^(?:( ){2,}|\t+)/;
var INDENT_TYPE_SPACE = " ";
var INDENT_TYPE_TAB = "	";
function encodeIndentsKey(indentType, indentAmount) {
  const typeCharacter = indentType === INDENT_TYPE_SPACE ? "s" : "t";
  return typeCharacter + String(indentAmount);
}
function decodeIndentsKey(indentsKey) {
  const keyHasTypeSpace = indentsKey[0] === "s";
  const type = keyHasTypeSpace ? INDENT_TYPE_SPACE : INDENT_TYPE_TAB;
  const amount = Number(indentsKey.slice(1));
  return { type, amount };
}
function getMostUsedKey(indents) {
  let result;
  let maxUsed = 0;
  let maxWeight = 0;
  for (const [key, [usedCount, weight]] of indents) {
    if (usedCount > maxUsed || usedCount === maxUsed && weight > maxWeight) {
      maxUsed = usedCount;
      maxWeight = weight;
      result = key;
    }
  }
  return result;
}
var IndentController = class {
  indents = /* @__PURE__ */ new Map();
  previousSize = 0;
  previousIndentType;
  key;
  constructor() {
  }
  recordIndent(value) {
    const matches = value.match(INDENT_REGEX);
    if (matches !== null) {
      const indent = matches[0].length;
      const indentType = matches[1] ? INDENT_TYPE_SPACE : INDENT_TYPE_TAB;
      if (indentType !== this.previousIndentType) {
        this.previousSize = 0;
      }
      this.previousIndentType = indentType;
      const indentDifference = indent - this.previousSize;
      this.previousSize = indent;
      if (indentDifference !== 0) {
        const absoluteIndentDifference = Math.abs(indentDifference);
        this.key = encodeIndentsKey(indentType, absoluteIndentDifference);
      }
      if (this.key) {
        const entry = this.indents.get(this.key) || [0, 0];
        this.indents.set(this.key, [entry[0] + 1, entry[1]]);
      }
    }
  }
  get output() {
    const keyOfMostUsedIndent = getMostUsedKey(this.indents);
    let type = "	";
    let amount = 1;
    if (keyOfMostUsedIndent !== void 0) {
      ;
      ({ type, amount } = decodeIndentsKey(keyOfMostUsedIndent));
    }
    return {
      type,
      amount
    };
  }
};

// src/tokenize-substrings.ts
function tokenizeSubstrings({
  value,
  start
}) {
  let state = "isQuote";
  const tokens = [];
  const buffer = [];
  let subStart = start;
  for (let i = 0; i < value.length; i++) {
    const char = value[i];
    if (["`", '"', "'"].includes(char)) {
      if (state !== "isQuote") {
        tokens.push({
          value: buffer.join(""),
          start: subStart,
          end: subStart + buffer.length
        });
        buffer.length = 0;
        subStart = i;
      }
      state = "isQuote";
    } else if (char.match(/^\s$/)) {
      if (state !== "isSpace") {
        tokens.push({
          value: buffer.join(""),
          start: subStart,
          end: subStart + buffer.length
        });
        buffer.length = 0;
        subStart = i;
      }
      state = "isSpace";
    } else {
      if (state !== "isChar") {
        tokens.push({
          value: buffer.join(""),
          start: subStart,
          end: subStart + buffer.length
        });
        buffer.length = 0;
        subStart = i;
      }
      state = "isChar";
    }
    buffer.push(char);
  }
  tokens.push({
    value: buffer.join(""),
    start: subStart,
    end: value.length
  });
  return tokens;
}

// src/tokenize-js.ts
import { Parser } from "acorn";
function tokenizeJavaScript(sourceCode) {
  const tokens = [];
  const tokenizer = Parser.tokenizer(sourceCode, {
    ecmaVersion: "latest",
    sourceType: "module"
  });
  const indentController = new IndentController();
  let token = tokenizer.getToken();
  while (token.type.label !== "eof") {
    const value = sourceCode.slice(token.start, token.end);
    if (token.type.label === "string") {
      const subtokens = tokenizeSubstrings({ value, start: token.start });
      tokens.push(...subtokens);
    } else {
      if (token.type.label === "whitespace") {
        indentController.recordIndent(value);
      }
      tokens.push({ value, start: token.start, end: token.end });
    }
    token = tokenizer.getToken();
  }
  return {
    tokens,
    indentType: indentController.output.type,
    indentAmount: indentController.output.amount
  };
}

// src/tokenize-ts.ts
import ts from "typescript";
function tokenizeTypeScript(sourceCode) {
  const tokens = [];
  const scanner = ts.createScanner(
    ts.ScriptTarget.Latest,
    false,
    ts.LanguageVariant.Standard,
    sourceCode
  );
  const indentController = new IndentController();
  let whileLimit = 1e4;
  let token = scanner.scan();
  while (token !== ts.SyntaxKind.EndOfFileToken && whileLimit-- >= 0) {
    const start = scanner.getTokenStart();
    const end = scanner.getTokenEnd();
    let value = sourceCode.slice(start, end);
    if (token === ts.SyntaxKind.WhitespaceTrivia) {
      indentController.recordIndent(value);
    }
    if (value.includes(" ")) {
      const subtokens = tokenizeSubstrings({ value, start });
      tokens.push(...subtokens);
    } else {
      tokens.push({ value, start, end });
    }
    token = scanner.scan();
  }
  if (whileLimit <= 0) {
    throw new Error("While limit reached");
  }
  return {
    tokens,
    indentType: indentController.output.type,
    indentAmount: indentController.output.amount
  };
}

// src/tokenize-txt.ts
function tokenizeText(sourceCode) {
  const tokens = [];
  const regex = /\S+/g;
  let match;
  while ((match = regex.exec(sourceCode)) !== null) {
    const value = match[0];
    const start = match.index;
    const end = start + value.length;
    tokens.push({ value, start, end });
  }
  return {
    tokens,
    indentType: " ",
    indentAmount: 1
  };
}

// src/tokenize.ts
function tokenize({
  content,
  language
}) {
  switch (language) {
    case "typescript":
      return tokenizeTypeScript(content);
    case "javascript":
      return tokenizeJavaScript(content);
    default:
      return tokenizeText(content);
  }
}

// src/stream.ts
var ReadableStreamWriter = class extends Writer {
  stream;
  controller = null;
  constructor() {
    super();
    this.stream = new ReadableStream({
      start: (controller) => {
        this.controller = controller;
      }
    });
  }
  write(operation) {
    if (this.controller) {
      this.controller.enqueue(operation);
    }
  }
  close() {
    if (this.controller) {
      this.controller.close();
    }
    return this.stream;
  }
};
function diffStreamToTextStream(diffStream, options = {}) {
  let omit = options.omit ?? [];
  let insertTagOpen = options.insertTagOpen ?? "[+ ";
  let insertTagClose = options.insertTagClose ?? " +]";
  let deleteTagOpen = options.deleteTagOpen ?? "[- ";
  let deleteTagClose = options.deleteTagClose ?? " -]";
  let equalTagOpen = options.equalTagOpen ?? "";
  let equalTagClose = options.equalTagClose ?? "";
  const encoder = new TextEncoder();
  const transformStream = new TransformStream({
    transform(change, controller) {
      let chunk = "";
      if (change.type === "insert" && !omit.includes("insert")) {
        chunk += insertTagOpen;
        for (const token of change.tokens) {
          chunk += token.value;
        }
        chunk += insertTagClose;
      }
      if (change.type === "delete" && !omit.includes("delete")) {
        chunk += deleteTagOpen;
        for (const token of change.tokens) {
          chunk += token.value;
        }
        chunk += deleteTagClose;
      }
      if (change.type === "equal" && !omit.includes("equal")) {
        chunk += equalTagOpen;
        for (const token of change.tokens) {
          chunk += token.value;
        }
        chunk += equalTagClose;
      }
      if (chunk) {
        controller.enqueue(encoder.encode(chunk));
      }
    }
  });
  return diffStream.pipeThrough(transformStream);
}
async function diffStreamToString(diffStream, options = {}) {
  const stream = diffStreamToTextStream(diffStream, options);
  const decoder = new TextDecoder();
  const reader = stream.getReader();
  let result = "";
  while (true) {
    const { done, value } = await reader.read();
    if (done) break;
    result += decoder.decode(value, { stream: true });
  }
  return result;
}
export {
  ArrayWriter,
  ReadableStreamWriter,
  diffArrayToString,
  diffStreamToString,
  diffStreamToTextStream,
  diffStringToArray,
  diffTokens,
  tokenize
};
//# sourceMappingURL=index.js.map